use std::path::PathBuf;

macro_rules! return_none {
    ($x:expr) => {
        match $x {
            Some(v) => v,
            None => return Ok(None),
        }
    };
}

pub struct Switches {
    args: Vec<String>,
}
impl Switches {
    pub fn new(args: Vec<String>) -> Self {
        Self { args }
    }

    pub fn from_env() -> Self {
        Self::new(std::env::args().collect())
    }

    pub fn get(&mut self, switch: &str) -> SwitchBuilder {
        SwitchBuilder::new(&mut self.args).or(switch)
    }

    pub fn or(&mut self, switch: &str) -> SwitchBuilder {
        self.get(switch)
    }

    pub fn check_for_unused_arguments(&mut self) -> Result<(), SwitchesError> {
        if !self.args.is_empty() {
            Err(SwitchesError::UnexpectedArguments(self.args.clone()))
        } else {
            Ok(())
        }
    }
}

pub struct SwitchBuilder<'a> {
    args: &'a mut Vec<String>,
    matched_indices: Vec<usize>,
    matched_switches: Vec<String>,
    all_switches: Vec<String>,
}

impl<'a> SwitchBuilder<'a> {
    fn new(args: &'a mut Vec<String>) -> Self {
        Self {
            args,
            matched_indices: Vec::new(),
            matched_switches: Vec::new(),
            all_switches: Vec::new(),
        }
    }

    pub fn or(mut self, switch: &str) -> Self {
        self.all_switches.push(switch.to_string());
        for (index, arg) in self.args.iter().enumerate() {
            if arg == switch {
                self.matched_indices.push(index);
                self.matched_switches.push(arg.to_string());
            }
        }
        self
    }

    pub fn as_flag(&mut self) -> Result<bool, SwitchesError> {
        let result = match self.matched_indices.len() {
            0 => Ok(false),
            1 => Ok(true),
            _ => Err(SwitchesError::MultipleMatchesFound(
                self.matched_switches.clone(),
            )),
        };
        self.remove_matches();
        result
    }

    pub fn as_opt_string(&mut self) -> Result<Option<String>, SwitchesError> {
        let switch_index = match self.matched_indices.len() {
            0 => return Ok(None),
            1 => self.matched_indices[0],
            _ => {
                return Err(SwitchesError::MultipleMatchesFound(
                    self.matched_switches.clone(),
                ))
            }
        };
        let switch = self.matched_switches[0].clone();
        let value_index = switch_index + 1;
        let value = match self.args.get(value_index) {
            Some(value) => {
                self.matched_indices.push(value_index);
                value.to_string()
            }
            None => return Err(SwitchesError::NoValueForSwitch(switch)),
        };
        if value.starts_with('-') {
            return Err(SwitchesError::NoValueForSwitch(switch));
        };
        self.remove_matches();
        Ok(Some(value))
    }

    pub fn as_opt_path(&mut self) -> Result<Option<PathBuf>, SwitchesError> {
        let string = return_none!(self.as_opt_string()?);
        let mut path = std::env::current_dir().unwrap();
        path.push(PathBuf::from(string));
        Ok(Some(path))
    }

    pub fn as_opt_u16(&mut self) -> Result<Option<u16>, SwitchesError> {
        let string = return_none!(self.as_opt_string()?);
        let int = string.parse().unwrap();
        Ok(Some(int))
    }

    pub fn as_string(&mut self) -> Result<String, SwitchesError> {
        let string = self.as_opt_string();
        self.make_mandatory(string)
    }
    pub fn as_path(&mut self) -> Result<PathBuf, SwitchesError> {
        let path = self.as_opt_path();
        self.make_mandatory(path)
    }
    pub fn as_u16(&mut self) -> Result<u16, SwitchesError> {
        let int = self.as_opt_u16();
        self.make_mandatory(int)
    }

    fn make_mandatory<T>(
        &mut self,
        result: Result<Option<T>, SwitchesError>,
    ) -> Result<T, SwitchesError> {
        match result? {
            Some(value) => Ok(value),
            None => Err(SwitchesError::NoMatchFound(self.all_switches.clone())),
        }
    }

    fn remove_matches(&mut self) {
        for index in self.matched_indices.iter().rev() {
            self.args.remove(*index);
        }
    }
}

fn join(vec: &[String]) -> String {
    let mut output = String::new();
    for (i, string) in vec.iter().enumerate() {
        match i == vec.len() - 1 {
            true => output.push_str(&format!("'{}'", string)),
            false => output.push_str(&format!("'{}', ", string)),
        }
    }
    output
}

pub enum SwitchesError {
    MultipleMatchesFound(Vec<String>), // contains all of the switches that were matched
    NoMatchFound(Vec<String>),         // contains all of the switches that were looked for
    NoValueForSwitch(String),
    UnexpectedArguments(Vec<String>),
    InvalidValue(String, String, String),
}
impl std::fmt::Display for SwitchesError {
    #[rustfmt::skip]
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        let string = match self {
            SwitchesError::MultipleMatchesFound(matches) => {
                format!("Multiple conflicting arguments were provided: {}", join(matches))}
            SwitchesError::NoMatchFound(switches) => {
                format!("One of these arguments must be provided: {}", join(switches))}
            SwitchesError::NoValueForSwitch(switch) => {
                format!("This argument needs to be followed by a value: '{}'", switch)}
            SwitchesError::UnexpectedArguments(args) => {
                format!("Unexpected arguments were provided: {}", join(args))}
            SwitchesError::InvalidValue(value, arg, ty) => {
                format!("A non-{} value was passed to '{}': {}", ty, arg, value)}

        };
        write!(f, "{}", string)
    }
}
